from cerillo import StratusReader


def test_Stratus():
    """Test a complete experiment setup workflow"""

    # Configure plate template
    reader = StratusReader(port="/dev/ttyUSB0")
    reader.connect()
    results = reader.start_endpoint_experiment(wavelength=600)
    print(results.to_json(indent=2))
    csv_long = results.to_csv(format='long')
    csv_wide = results.to_csv(format='wide')

    # Save to disk
    results.save_json('experiment.json')
    results.save_csv('experiment_long.csv', format='long')
    results.save_csv('experiment_wide.csv', format='wide')



test_Stratus()



def index_to_well(index: int, n_cols=12) -> str:
    """
    Convert a 0..95 index into a 96-well plate label (A1..H12).

    Parameters
    ----------
    index : int
        Integer index in range 0..95.

    Returns
    -------
    str
        Well label like "A1", "H12".

    Raises
    ------
    ValueError
        If index is out of range or `order` is invalid.
    """
    if not isinstance(index, int):
        raise TypeError("index must be an integer")
    if index < 0 or index > 95:
        raise ValueError("index must be between 0 and 95 inclusive")

    rows = "ABCDEFGH"
    row_idx = index // n_cols
    col_idx = index % n_cols

    return f"{rows[row_idx]}{col_idx + 1}"



# src/opentrons_plate_reader/experiment_builder.py
"""
Builder classes for constructing Stratus experiments with fluent API
"""

from typing import Optional, List, Tuple
from dataclasses import dataclass, field
from .cdcl import CDCLCommand
from .start_commands import (
    ResetStartCommand,
    SetIntervalCommand,
    SetDurationCommand,
    SetModeCommand,
    SetExperimentNameCommand,
    SetExperimentIDCommand,
    SetUserCommand,
    SetExperimentTypeCommand,
    SetQueueCommand,
    ExecuteStartCommand,
    SetPlateRowsCommand,
    SetPlateColsCommand,
    SetWellDiameterCommand,
    SetWellHeightCommand,
    SetWellVolumeCommand,
    SetPlateTreatmentCommand,
    SetWellBottomCommand,
    SetPlateLidCommand,
    AssignDetectorToWellCommand,
    SetPlateManufacturerCommand,
    SetPlateModelCommand,
    SetPlateTemplateNameCommand, AddPlateLayoutLineCommand
)
from .codes import (
    ExperimentType,
    PlateTreatmentType,
    WellBottomType,
    PlateLidType
)


class ExperimentBuilder:
    """
    Fluent builder for constructing Stratus experiment configurations.

    Example:
        builder = ExperimentBuilder()
        commands = (builder
            .reset()
            .set_interval(60)
            .set_duration(3600)
            .set_name("My Experiment")
            .set_user("researcher@lab.com")
            .set_type(ExperimentType.KINETIC)
            .build())
    """

    def __init__(self, command_id: Optional[str] = None):
        """
        Initialize experiment builder.

        Args:
            command_id: Optional command ID to use for all commands
        """
        self._commands: List[CDCLCommand] = []
        self._command_id = command_id

    def reset(self) -> 'ExperimentBuilder':
        """Reset start command to defaults"""
        cmd = ResetStartCommand(command_id=self._command_id)
        self._commands.append(cmd)
        return self

    def set_interval(self, interval: int) -> 'ExperimentBuilder':
        """
        Set measurement interval in seconds.

        Args:
            interval: Interval in seconds (-1 to ignore)
        """
        cmd = SetIntervalCommand(
            interval=interval, command_id=self._command_id)
        self._commands.append(cmd)
        return self

    def set_duration(self, duration: int) -> 'ExperimentBuilder':
        """
        Set measurement duration in seconds.

        Args:
            duration: Duration in seconds (-1 for infinite)
        """
        cmd = SetDurationCommand(
            duration=duration, command_id=self._command_id)
        self._commands.append(cmd)
        return self

    def set_mode(self, mode: int) -> 'ExperimentBuilder':
        """
        Set measurement mode bitfield.

        Args:
            mode: Mode bitfield value
        """
        cmd = SetModeCommand(mode=mode, command_id=self._command_id)
        self._commands.append(cmd)
        return self

    def set_name(self, name: str) -> 'ExperimentBuilder':
        """
        Set experiment name.

        Args:
            name: Experiment name (empty to auto-generate)
        """
        cmd = SetExperimentNameCommand(name=name, command_id=self._command_id)
        self._commands.append(cmd)
        return self

    def set_experiment_id(self, experiment_id: str) -> 'ExperimentBuilder':
        """
        Set experiment ID.

        Args:
            experiment_id: UUID v4 string (empty to auto-generate)
        """
        cmd = SetExperimentIDCommand(
            experiment_id=experiment_id,
            command_id=self._command_id
        )
        self._commands.append(cmd)
        return self

    def set_user(self, user: str) -> 'ExperimentBuilder':
        """
        Set user who started the experiment.

        Args:
            user: User identifier (max 64 characters)
        """
        cmd = SetUserCommand(user=user, command_id=self._command_id)
        self._commands.append(cmd)
        return self

    def set_type(self, exp_type: ExperimentType) -> 'ExperimentBuilder':
        """
        Set experiment type.

        Args:
            exp_type: ExperimentType enum value
        """
        cmd = SetExperimentTypeCommand(
            exp_type=exp_type, command_id=self._command_id)
        self._commands.append(cmd)
        return self

    def set_queue(self, queue: bool = True) -> 'ExperimentBuilder':
        """
        Set whether to queue the experiment.

        Args:
            queue: True to queue, False to start immediately
        """
        cmd = SetQueueCommand(queue=queue, command_id=self._command_id)
        self._commands.append(cmd)
        return self

    def build(self) -> List[CDCLCommand]:
        """
        Build and return the list of commands.

        Returns:
            List of CDCLCommand objects ready to send
        """
        commands = self._commands.copy()
        # Ensure experiment commands include a reset at the start for safety
        if not any(isinstance(c, ResetStartCommand) for c in commands):
            commands = [ResetStartCommand(
                command_id=self._command_id)] + commands

        # Ensure an execute is present at the end so messages built via this
        # convenience method will start the experiment automatically.
        if not isinstance(commands[len(commands)-1], ExecuteStartCommand):
            commands = commands + \
                [ExecuteStartCommand(
                    command_id=self._command_id)]
        return commands

    def build_messages(self) -> List[str]:
        """
        Build and return raw message strings.

        Returns:
            List of raw CDCL message strings
        """
        # Make a shallow copy so we don't mutate internal state
        commands = self.build()

        return [cmd.raw_message for cmd in commands]

    def clear(self) -> 'ExperimentBuilder':
        """Clear all commands from the builder"""
        self._commands.clear()
        return self


class PlateTemplateBuilder:
    """
    Fluent builder for constructing plate templates.

    Example:
        builder = PlateTemplateBuilder()
        commands = (builder
            .set_96_well_plate()
            .set_manufacturer("Corning")
            .set_model("3595")
            .set_flat_bottom()
            .set_tissue_culture_treated()
            .build())
    """

    def __init__(self, command_id: Optional[str] = None):
        """
        Initialize plate template builder.

        Args:
            command_id: Optional command ID to use for all commands
        """
        self._commands: List[CDCLCommand] = []
        self._command_id = command_id

    def set_rows(self, rows: int) -> 'PlateTemplateBuilder':
        """Set number of rows"""
        cmd = SetPlateRowsCommand(rows=rows, command_id=self._command_id)
        self._commands.append(cmd)
        return self

    def set_cols(self, cols: int) -> 'PlateTemplateBuilder':
        """Set number of columns"""
        cmd = SetPlateColsCommand(cols=cols, command_id=self._command_id)
        self._commands.append(cmd)
        return self

    def set_well_diameter(self, diameter_um: int) -> 'PlateTemplateBuilder':
        """Set well diameter in microns"""
        cmd = SetWellDiameterCommand(
            diameter_um=diameter_um, command_id=self._command_id)
        self._commands.append(cmd)
        return self

    def set_well_height(self, height_um: int) -> 'PlateTemplateBuilder':
        """Set well height in microns"""
        cmd = SetWellHeightCommand(
            height_um=height_um, command_id=self._command_id)
        self._commands.append(cmd)
        return self

    def set_well_volume(self, volume_ul: int) -> 'PlateTemplateBuilder':
        """Set well volume in microliters"""
        cmd = SetWellVolumeCommand(
            volume_ul=volume_ul, command_id=self._command_id)
        self._commands.append(cmd)
        return self

    def set_treatment(self, treatment: PlateTreatmentType) -> 'PlateTemplateBuilder':
        """Set plate treatment type"""
        cmd = SetPlateTreatmentCommand(
            treatment=treatment, command_id=self._command_id)
        self._commands.append(cmd)
        return self

    def set_tissue_culture_treated(self) -> 'PlateTemplateBuilder':
        """Set plate as tissue-culture treated"""
        return self.set_treatment(PlateTreatmentType.TISSUE_CULTURE)

    def set_untreated(self) -> 'PlateTemplateBuilder':
        """Set plate as untreated"""
        return self.set_treatment(PlateTreatmentType.UNTREATED)

    def set_bottom_type(self, bottom: WellBottomType) -> 'PlateTemplateBuilder':
        """Set well bottom type"""
        cmd = SetWellBottomCommand(bottom=bottom, command_id=self._command_id)
        self._commands.append(cmd)
        return self

    def set_flat_bottom(self) -> 'PlateTemplateBuilder':
        """Set wells as flat-bottom"""
        return self.set_bottom_type(WellBottomType.FLAT)

    def set_round_bottom(self) -> 'PlateTemplateBuilder':
        """Set wells as round-bottom"""
        return self.set_bottom_type(WellBottomType.ROUND)

    def set_lid_type(self, lid: PlateLidType) -> 'PlateTemplateBuilder':
        """Set plate lid type"""
        cmd = SetPlateLidCommand(lid=lid, command_id=self._command_id)
        self._commands.append(cmd)
        return self

    def set_standard_lid(self) -> 'PlateTemplateBuilder':
        """Set standard lid"""
        return self.set_lid_type(PlateLidType.LID)

    def set_no_lid(self) -> 'PlateTemplateBuilder':
        """Set no lid"""
        return self.set_lid_type(PlateLidType.NONE)

    def set_breathe_easy(self) -> 'PlateTemplateBuilder':
        """Set permeable membrane (e.g., Breathe-Easy)"""
        return self.set_lid_type(PlateLidType.PERMEABLE_MEMBRANE)

    def assign_detector_to_well(
        self,
        detector_num: int,
        well_num: int
    ) -> 'PlateTemplateBuilder':
        """
        Assign a detector to measure a specific well.

        Args:
            detector_num: Detector number
            well_num: Well number
        """
        cmd = AssignDetectorToWellCommand(
            detector_num=detector_num,
            well_num=well_num,
            command_id=self._command_id
        )
        self._commands.append(cmd)
        return self

    def set_manufacturer(self, manufacturer: str) -> 'PlateTemplateBuilder':
        """Set plate manufacturer"""
        cmd = SetPlateManufacturerCommand(
            manufacturer=manufacturer,
            command_id=self._command_id
        )
        self._commands.append(cmd)
        return self

    def set_model(self, model: str) -> 'PlateTemplateBuilder':
        """Set plate model number"""
        cmd = SetPlateModelCommand(model=model, command_id=self._command_id)
        self._commands.append(cmd)
        return self

    def set_template_name(self, name: str) -> 'PlateTemplateBuilder':
        """Set plate template name"""
        cmd = SetPlateTemplateNameCommand(
            name=name, command_id=self._command_id)
        self._commands.append(cmd)
        return self

    def add_plate_layout_line(self, line: str) -> 'PlateTemplateBuilder':
        """Add to plate layout file"""
        cmd = AddPlateLayoutLineCommand(line=line, command_id=self._command_id)
        self._commands.append(cmd)
        return self

    # Convenience methods for common plate types
    def set_96_well_plate(self) -> 'PlateTemplateBuilder':
        """Configure as standard 96-well plate"""
        return (self
                .set_rows(8)
                .set_cols(12)
                .set_well_diameter(6400)
                .set_well_height(10900)
                .set_well_volume(200))

    def set_24_well_plate(self) -> 'PlateTemplateBuilder':
        """Configure as standard 24-well plate"""
        return (self
                .set_rows(4)
                .set_cols(6)
                .set_well_diameter(15500)
                .set_well_height(17400)
                .set_well_volume(340))

    def set_6_well_plate(self) -> 'PlateTemplateBuilder':
        """Configure as standard 6-well plate"""
        return (self
                .set_rows(2)
                .set_cols(3)
                .set_well_diameter(35000)
                .set_well_height(18500)
                .set_well_volume(1500))

    def set_12_well_plate(self) -> 'PlateTemplateBuilder':
        """Configure as standard 12-well plate"""
        return (self
                .set_rows(3)
                .set_cols(4)
                .set_well_diameter(22800)
                .set_well_height(17000)
                .set_well_volume(852))

    def build(self) -> List[CDCLCommand]:
        """
        Build and return the list of commands.

        Returns:
            List of CDCLCommand objects ready to send
        """
        return self._commands.copy()

    def build_messages(self) -> List[str]:
        """
        Build and return raw message strings.

        Returns:
            List of raw CDCL message strings
        """
        return [cmd.raw_message for cmd in self._commands]

    def clear(self) -> 'PlateTemplateBuilder':
        """Clear all commands from the builder"""
        self._commands.clear()
        return self


class FullExperimentBuilder:
    """
    Combined builder for both experiment settings and plate template.

    Example:
        builder = FullExperimentBuilder()
        builder
            .experiment()
                .set_name("Growth Curve")
                .set_interval(300)
                .set_duration(86400)
            builder
            .plate()
                .set_96_well_plate()
                .set_manufacturer("Corning")
            commands = builder.build_all()
    """

    def __init__(self, command_id: Optional[str] = None):
        """
        Initialize full experiment builder.

        Args:
            command_id: Optional command ID to use for all commands
        """
        self._experiment_builder = ExperimentBuilder(command_id)
        self._plate_builder = PlateTemplateBuilder(command_id)
        self._current_builder = self._experiment_builder

    def experiment(self) -> ExperimentBuilder:
        """
        Switch to experiment configuration.

        Returns:
            ExperimentBuilder for chaining
        """
        self._current_builder = self._experiment_builder
        return self._experiment_builder

    def plate(self) -> PlateTemplateBuilder:
        """
        Switch to plate template configuration.

        Returns:
            PlateTemplateBuilder for chaining
        """
        self._current_builder = self._plate_builder
        return self._plate_builder

    def build_all(self) -> List[CDCLCommand]:
        """
        Build all commands (plate template first, then experiment).

        Returns:
            Combined list of all commands
        """
        plate_commands = self._plate_builder.build()
        exp_commands = self._experiment_builder.build()
        # strip off the exp end command ( we append it later)
        if isinstance(exp_commands[len(exp_commands)-1], ExecuteStartCommand):
            exp_commands = exp_commands[0:len(exp_commands)-1]
        full_commands = exp_commands + plate_commands

        # Ensure experiment commands include a reset at the start for safety
        if not any(isinstance(c, ResetStartCommand) for c in exp_commands):
            exp_commands = [ResetStartCommand(
                command_id=self._experiment_builder._command_id)] + exp_commands

        if not isinstance(full_commands[len(full_commands)-1], ExecuteStartCommand):
            full_commands = full_commands + \
                [ExecuteStartCommand(
                    command_id=self._experiment_builder._command_id)]

        return full_commands

    def build_all_messages(self) -> List[str]:
        """
        Build all raw message strings.

        Returns:
            Combined list of all message strings
        """
        # Build command objects but avoid mutating the builders' internal lists
        full_commands = self.build_all()

        # Ensure an execute is present at the end so messages built via this
        # convenience method will start the experiment automatically.
        return [cmd.raw_message for cmd in full_commands]



# tests/test_experiment_builder.py
import pytest
from cerillo.cdcl import (
    ExperimentBuilder,
    PlateTemplateBuilder,
    FullExperimentBuilder,
    ExperimentType,
    PlateTreatmentType,
    WellBottomType,
    PlateLidType
)
from cerillo.experiment_results import (
    ExperimentResults, DataTransmissionParser)


class TestExperimentBuilder:
    """Test suite for ExperimentBuilder"""

    def test_simple_build(self):
        """Test building a simple experiment"""
        builder = ExperimentBuilder()
        commands = builder.set_interval(60).build()

        assert len(commands) == 3
        assert commands[1].raw_message == "A i 60"

    def test_chained_commands(self):
        """Test chaining multiple commands"""
        builder = ExperimentBuilder()
        commands = (builder
                    .reset()
                    .set_interval(60)
                    .set_duration(3600)
                    .set_name("Test")
                    .build())

        assert len(commands) == 5
        assert commands[0].raw_message == "A 0"
        assert commands[1].raw_message == "A i 60"
        assert commands[2].raw_message == "A d 3600"
        assert commands[3].raw_message == "A n Test"
        assert commands[4].raw_message == "A ."

    def test_experiment_types(self):
        """Test setting different experiment types"""
        builder = ExperimentBuilder()

        # Kinetic
        commands = builder.set_type(ExperimentType.KINETIC).build()
        assert commands[1].raw_message == "A t 1"

        builder.clear()

        # Endpoint
        commands = builder.set_type(ExperimentType.ENDPOINT).build()
        assert commands[1].raw_message == "A t 0"

    def test_queue_command(self):
        """Test queue command"""
        builder = ExperimentBuilder()

        # Queue = True
        commands = builder.set_queue(True).build()
        assert commands[1].raw_message == "A q 1"

        builder.clear()

        # Queue = False
        commands = builder.set_queue(False).build()
        assert commands[1].raw_message == "A q 0"

    def test_with_command_id(self):
        """Test using command IDs"""
        builder = ExperimentBuilder(command_id="test")
        commands = builder.set_interval(60).build()

        assert len(commands) == 3
        assert commands[0].command_id == "test"
        assert commands[1].raw_message == "A*test i 60"

    def test_build_messages(self):
        """Test build_messages returns raw strings"""
        builder = ExperimentBuilder()
        messages = (builder
                    .set_interval(60)
                    .set_duration(3600)
                    .build_messages())

        assert len(messages) == 4
        assert isinstance(messages[0], str)
        assert messages[0] == 'A 0'
        assert messages[1] == "A i 60"
        assert messages[2] == "A d 3600"
        assert messages[3] == 'A .'

    def test_clear(self):
        """Test clearing builder"""
        builder = ExperimentBuilder()
        builder.set_interval(60).set_duration(3600)

        assert len(builder.build()) == 4

        builder.clear()
        assert len(builder.build()) == 2

    def test_user_command(self):
        """Test setting user"""
        builder = ExperimentBuilder()
        commands = builder.set_user("researcher@lab.com").build()

        assert commands[1].raw_message == "A U researcher@lab.com"

    def test_experiment_id(self):
        """Test setting experiment ID"""
        builder = ExperimentBuilder()
        commands = builder.set_experiment_id("uuid-1234").build()

        assert commands[1].raw_message == "A I uuid-1234"


class TestPlateTemplateBuilder:
    """Test suite for PlateTemplateBuilder"""

    def test_96_well_plate(self):
        """Test 96-well plate preset"""
        builder = PlateTemplateBuilder()
        commands = builder.set_96_well_plate().build()

        # Should set 5 parameters (rows, cols, diameter, height, volume)
        assert len(commands) == 5

        messages = [cmd.raw_message for cmd in commands]
        assert "A PMR 8" in messages
        assert "A PMC 12" in messages

    def test_6_well_plate(self):
        """Test 6-well plate preset"""
        builder = PlateTemplateBuilder()
        commands = builder.set_6_well_plate().build()

        messages = [cmd.raw_message for cmd in commands]
        assert "A PMR 2" in messages
        assert "A PMC 3" in messages

    def test_custom_dimensions(self):
        """Test setting custom dimensions"""
        builder = PlateTemplateBuilder()
        commands = (builder
                    .set_rows(4)
                    .set_cols(6)
                    .set_well_diameter(10000)
                    .build())

        assert len(commands) == 3
        assert commands[0].raw_message == "A PMR 4"
        assert commands[1].raw_message == "A PMC 6"
        assert commands[2].raw_message == "A PMD 10000"

    def test_treatment_types(self):
        """Test setting plate treatment"""
        builder = PlateTemplateBuilder()

        # Tissue culture treated
        commands = builder.set_tissue_culture_treated().build()
        assert commands[0].raw_message == "A PMT t"

        builder.clear()

        # Untreated
        commands = builder.set_untreated().build()
        assert commands[0].raw_message == "A PMT u"

    def test_bottom_types(self):
        """Test setting well bottom types"""
        builder = PlateTemplateBuilder()

        # Flat bottom
        commands = builder.set_flat_bottom().build()
        assert commands[0].raw_message == "A PMB f"

        builder.clear()

        # Round bottom
        commands = builder.set_round_bottom().build()
        assert commands[0].raw_message == "A PMB r"

    def test_lid_types(self):
        """Test setting lid types"""
        builder = PlateTemplateBuilder()

        # Standard lid
        commands = builder.set_standard_lid().build()
        assert commands[0].raw_message == "A PML l"

        builder.clear()

        # No lid
        commands = builder.set_no_lid().build()
        assert commands[0].raw_message == "A PML n"

        builder.clear()

        # Breathe-Easy
        commands = builder.set_breathe_easy().build()
        assert commands[0].raw_message == "A PML m"

    def test_detector_assignment(self):
        """Test assigning detectors to wells"""
        builder = PlateTemplateBuilder()
        commands = builder.assign_detector_to_well(1, 0).build()

        assert len(commands) == 1
        assert commands[0].raw_message == "A PMW 1 0"

    def test_metadata(self):
        """Test setting plate metadata"""
        builder = PlateTemplateBuilder()
        commands = (builder
                    .set_manufacturer("Corning")
                    .set_model("3595")
                    .set_template_name("Test Plate")
                    .build())

        assert len(commands) == 3
        assert commands[0].raw_message == "A PMM Corning"
        assert commands[1].raw_message == "A PMm 3595"
        assert commands[2].raw_message == "A PMn Test Plate"

    def test_complete_plate_config(self):
        """Test complete plate configuration"""
        builder = PlateTemplateBuilder()
        messages = (builder
                    .set_96_well_plate()
                    .set_manufacturer("Corning")
                    .set_model("3595")
                    .set_flat_bottom()
                    .set_tissue_culture_treated()
                    .set_standard_lid()
                    .build_messages())

        assert len(messages) == 10
        assert "A PMR 8" in messages
        assert "A PMC 12" in messages
        assert "A PMM Corning" in messages
        assert "A PMm 3595" in messages
        assert "A PMB f" in messages
        assert "A PMT t" in messages
        assert "A PML l" in messages


class TestFullExperimentBuilder:
    """Test suite for FullExperimentBuilder"""

    def test_experiment_and_plate(self):
        """Test building both experiment and plate"""
        builder = FullExperimentBuilder()

        builder.plate().set_96_well_plate()
        builder.experiment().set_interval(60)

        commands = builder.build_all()

        # Should have plate commands first, then experiment commands
        assert len(commands) > 5

        # First commands should be plate template (A P M ...)
        messages = [cmd.raw_message for cmd in commands]
        plate_msgs = [m for m in messages if m.startswith("A P")]
        exp_msgs = [m for m in messages if m.startswith("A i")]

        assert len(plate_msgs) == 5  # 96-well preset has 5 commands
        assert len(exp_msgs) == 1

    def test_build_all_messages(self):
        """Test building all messages as strings"""
        builder = FullExperimentBuilder()

        builder.plate().set_96_well_plate()
        builder.experiment().set_interval(60)

        messages = builder.build_all_messages()

        assert len(messages) == 8  # 5 plate + 3 experiment
        assert all(isinstance(m, str) for m in messages)

    def test_switching_contexts(self):
        """Test switching between plate and experiment contexts"""
        builder = FullExperimentBuilder()

        # Configure plate
        plate_builder = builder.plate()
        plate_builder.set_96_well_plate()

        # Switch to experiment
        exp_builder = builder.experiment()
        exp_builder.set_interval(60)

        # Switch back to plate
        plate_builder2 = builder.plate()
        plate_builder2.set_manufacturer("Test")

        # Should be same builder
        assert plate_builder is plate_builder2

        commands = builder.build_all()
        assert len(commands) == 9  # 5 + 1 plate, 1 experiment

    def test_complete_workflow(self):
        """Test a complete experiment setup workflow"""
        builder = FullExperimentBuilder(command_id="test")

        # Configure plate template
        builder.plate().set_96_well_plate()
        builder.plate().set_manufacturer("Greiner")
        builder.plate().set_model("655101")
        builder.plate().set_flat_bottom()
        builder.plate().set_tissue_culture_treated()

        # Configure experiment
        builder.experiment().set_name("Growth Curve")
        builder.experiment().set_user("scientist@lab.com")
        builder.experiment().set_interval(300)
        builder.experiment().set_duration(86400)
        builder.experiment().set_type(ExperimentType.KINETIC)

        messages = builder.build_all_messages()

        # Check that all commands have the command ID
        assert all("*test " in m for m in messages)

        # Check we have both plate and experiment commands
        plate_count = sum(1 for m in messages if "A*test PM" in m)
        exp_count = sum(1 for m in messages if m.startswith(
            "A*test") and "P" not in m)

        assert plate_count == 9
        assert exp_count == 7   # reset + 5 settings + execute


class TestBuilderEdgeCases:
    """Test edge cases and error conditions"""

    def test_empty_builder(self):
        """Test building with no commands"""
        builder = ExperimentBuilder()
        commands = builder.build()

        assert len(commands) == 2
        assert commands[0].raw_message == 'A 0'
        assert commands[1].raw_message == 'A .'

    def test_multiple_builds(self):
        """Test that build() doesn't clear commands"""
        builder = ExperimentBuilder()
        builder.set_interval(60)

        commands1 = builder.build()
        commands2 = builder.build()

        assert len(commands1) == 3
        assert len(commands2) == 3
        assert commands1[1].raw_message == commands2[1].raw_message

    def test_build_returns_copy(self):
        """Test that build() returns a copy of commands"""
        builder = ExperimentBuilder()
        builder.set_interval(60)

        commands1 = builder.build()
        commands2 = builder.build()

        # Should be different list objects
        assert commands1 is not commands2

        # But contain equivalent commands
        assert len(commands1) == len(commands2)

    def test_infinite_duration(self):
        """Test setting infinite duration"""
        builder = ExperimentBuilder()
        commands = builder.set_duration(-1).build()

        assert commands[1].raw_message == "A d -1"

    def test_empty_name(self):
        """Test empty experiment name (auto-generate)"""
        builder = ExperimentBuilder()
        commands = builder.set_name("").build()

        # Should still create command but without name parameter
        assert commands[1].raw_message == "A n"

    def test_empty_experiment_id(self):
        """Test empty experiment ID (auto-generate)"""
        builder = ExperimentBuilder()
        commands = builder.set_experiment_id("").build()

        assert commands[1].raw_message == "A I"

    def test_detector_assignment_multiple_wells(self):
        """Test assigning multiple detectors to wells"""
        builder = PlateTemplateBuilder()

        # Assign 3 detectors to 3 wells
        for i in range(3):
            builder.assign_detector_to_well(i + 1, i)

        commands = builder.build()
        assert len(commands) == 3

        messages = [cmd.raw_message for cmd in commands]
        assert "A PMW 1 0" in messages
        assert "A PMW 2 1" in messages
        assert "A PMW 3 2" in messages


class TestBuilderIntegration:
    """Integration tests with realistic scenarios"""

    def test_kinetic_growth_curve(self):
        """Test setting up a typical bacterial growth curve"""
        builder = FullExperimentBuilder()
        (builder
            .experiment()
            .reset()
            .set_name("E. coli Growth Curve")
            .set_user("microbiologist@lab.edu")
            .set_interval(600)      # 10 minutes
            .set_duration(43200)    # 12 hours
            .set_type(ExperimentType.KINETIC)
         )

        (builder
         .plate()
         .set_96_well_plate()
         .set_manufacturer("Corning")
         .set_flat_bottom()
         .set_tissue_culture_treated())
        builder.experiment()

        messages = (
            builder.build_all_messages())

        assert len(messages) > 10
        assert any("Growth Curve" in m for m in messages)
        assert any("A t 1" in m for m in messages)  # Kinetic type

    def test_endpoint_elisa(self):
        """Test setting up an ELISA endpoint read"""
        builder = FullExperimentBuilder()

        builder.plate().set_96_well_plate().set_manufacturer("Thermo").set_flat_bottom()
        builder.experiment().reset().set_name("ELISA Plate 1").set_type(
            ExperimentType.ENDPOINT)
        messages = (builder.build_all_messages())

        assert any("ELISA" in m for m in messages)
        assert any("A t 0" in m for m in messages)  # Endpoint type

    def test_6_well_cell_culture(self):
        """Test 6-well plate for cell culture"""
        builder = FullExperimentBuilder()

        # Build plate with detector assignments
        plate = builder.plate()
        plate.set_6_well_plate()
        plate.set_tissue_culture_treated()

        # Assign detectors (example: 4 detectors per well for 6 wells)
        for well in range(6):
            for detector in range(4):
                plate.assign_detector_to_well(
                    detector_num=well * 4 + detector + 1,
                    well_num=well
                )

        messages = builder.build_all_messages()

        # Should have 6-well config + 24 detector assignments
        detector_msgs = [m for m in messages if "A PMW" in m]
        assert len(detector_msgs) == 24


class TestExperimentResults:
    def test_basic_case(self):

        # Create results container
        results = ExperimentResults(
            experiment_id="A3C14980-B645-4D8D-8913-DF4231D32B95",
            experiment_name="Growth Curve Test",
            wavelengths=[450],
            plate_layout={'rows': 8, 'cols': 12}
        )

        message = ':12345*abcd D A3C14980-B645-4D8D-8913-DF4231D32B95\t1654625593\t28.5313\t262153\t0.0175\t0.0149\t0.0203\t0.0097\t0.0219\t0.0110\t0.0082\t0.0110\t0.0168\t0.0097\t0.0160\t0.0100\t0.0172\t0.0084\t0.0234\t0.0119\t0.0163\t0.0125\t0.0166\t0.0144\t0.0199\t0.0132\t0.0194\t0.0096\t0.0165\t0.0098\t0.0177\t0.0095\t0.0155\t0.0126\t0.0167\t0.0124\t0.0168\t0.0120\t0.0172\t0.0085\t0.0215\t0.0118\t0.0157\t0.0109\t0.0186\t0.0093\t0.0204\t0.0095\t0.0176\t0.0094\t0.0190\t0.0103\t0.0168\t0.0096\t0.0203\t0.0136\t0.0197\t0.0108\t0.0173\t0.0094\t0.0158\t0.0084\t0.0161\t0.0095\t0.0180\t0.0099\t0.0170\t0.0125\t0.0183\t0.0110\t0.0228\t0.0099\t0.0203\t0.0129\t0.0159\t0.0102\t0.0193\t0.0096\t0.0214\t0.0099\t0.0167\t0.0123\t0.0181\t0.0117\t0.0019\t0.0089\t0.0018\t0.0100\t0.0164\t0.0108\t0.0208\t0.0094\t0.0158\t0.0153\t0.0262\t0.0079\t0.0169\t0.0095\t0.0199\t0.0081'
        DataTransmissionParser.parse_data_message(message, results)

        # Simulate adding data points
        print(results)
        print(f"\nWells: {results.get_wells()}")
        print(f"\nTimeseries for A1:")
        for ts, val in results.get_well_timeseries('A1'):
            print(f'  {ts}: {float(val):.3f}')

        # Export examples
        print("\n=== JSON Export ===")
        print(results.to_json())

        print("\n=== CSV Export (long) ===")
        print(results.to_csv(format="long")[:300] + "...")

        print("\n=== CSV Export (wide) ===")
        print(results.to_csv(format="wide")[:300] + "...")

    def test_real_case(self):
        message = ':2000232 D"39034871-4198-4686-A58B-936B265F3B05"   1764083902      22.9277 0               -0.0289 -0.0366 -0.0165 0.0363  -0.0266 -0.0461 -0.0282 -0.0039 -0.0703 -0.0297 0.0667  0.0221    -0.0692 -0.0273 -0.0476 0.0192  -0.0123 -0.0816 -0.0372 -0.0243 0.0100  -0.0152 0.0232  0.0453  -0.0405 -0.0159 0.0414  -0.0120 -0.0088 0.0522  -0.0039 -0.0088 -0.0226 -0.0004 -0.0773   0.0397  -0.0636 -0.0468 -0.0638 -0.0613 0.0063  -0.0175 -0.0195 0.0063  -0.0360 -0.0387 -0.0392 -0.0519 -0.0707 -0.0637 -0.0608 -27.9529        -INF    -4.3024 -INF    -4.3980 -INF      -INF    -28.2724        19.2329 -INF    -INF    -INF    5.2151  -INF    -INF    -INF    20.3716 NaN     -27.6429        -INF    -INF    NaN     -4.1936 20.4781 -5.5370 -INF    -INF      1.1772  NaN     NaN     13.7477 -5.2987 -INF    -INF    -4.8702 -5.4781 NaN     NaN     NaN     -INF    -INF    NaN     -9.7185 13.8419 14.9864'
        # Create results container
        results = ExperimentResults(
            experiment_id="39034871-4198-4686-A58B-936B265F3B05",
            experiment_name="Growth Curve Test",
            wavelengths=[450],
            plate_layout={'rows': 8, 'cols': 12}
        )
        DataTransmissionParser.parse_data_message(message, results)

        # Simulate adding data points
        print(results)


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



from dataclasses import dataclass
from typing import Optional, List, Any
import re


@dataclass
class CDCLResponse:
    """Parsed CDCL protocol response"""
    serial_number: int
    command_id: Optional[str]
    prefix: str
    params: List[str]
    raw_message: str

    def __repr__(self):
        cmd_str = f" cmd_id={self.command_id}" if self.command_id else ""
        return (f"CDCLResponse(serial={self.serial_number}{cmd_str}, "
                f"prefix='{self.prefix}', params={self.params})")

    def __post_init__(self):
        """Generate raw_message if not provided"""
        if not self.raw_message:
            self.raw_message = self.to_message()


@dataclass
class CDCLCommand:
    """parsed CDCL protocol command"""
    command: str
    command_id: Optional[str] = None
    params: Optional[List[str]] = None
    raw_message: str = ""

    def __repr__(self):
        cmd_str = f" command_id={self.command_id}" if self.command_id else ""
        return (f"CDCLCommand({cmd_str}, "
                f"command='{self.command}', params={self.params})")

    def __post_init__(self):
        """Generate raw_message if not provided"""
        if not self.raw_message:
            self.raw_message = self.to_message()

    def to_message(self) -> str:
        """Construct the raw CDCL message from components"""
        # Start with serial number
        # Add prefix
        msg = f"{self.command}"

        # Add command ID if present
        if self.command_id:
            msg += f"*{self.command_id}"

        # Add params if present
        if self.params:
            msg += " " + " ".join(str(p) for p in self.params)

        return msg


class CDCLParseError(Exception):
    """Exception raised when parsing CDCL message fails"""
    pass


class CDCLParser:
    """Parser for CDCL protocol messages"""

    # Regex pattern for CDCL response message
    # Format: :<serial>*[<cmd_id>*]<prefix> [params...]
    CDCL_RESPONSE_PATTERN = re.compile(
        r'^:'                                 # Colon at start
        # Serial number (group 1)
        r'([A-Za-z0-9]+)'
        # Optional 4-char base64 cmd ID (group 2)
        r'(?:\*([A-Za-z0-9\+\/]{4}))?\s'
        # Single char prefix (group 3)
        r'([A-Za-z0-9])'
        # Optional space and params (group 4)
        r'\s*(?:(.*))?$'
    )

    CDCL_COMMAND_PATTERN = re.compile(
        r'^'
        r'([\S]{1})'
        r'(?:\*([A-Za-z0-9\+\/]{4}))?'
        r'\s*(?:(.*))?$'
    )

    @classmethod
    def parse_response(cls, message: str) -> CDCLResponse:
        """
        Parse a CDCL protocol message.

        Args:
            message: Raw CDCL message string

        Returns:
            CDCLResponse object with parsed fields

        Raises:
            CDCLParseError: If message doesn't match CDCL format
        """
        if not message:
            raise CDCLParseError("Empty message")

        # Strip whitespace
        message = message.strip()

        # Check for colon prefix
        if not message.startswith(':'):
            raise CDCLParseError(
                f"Response message must start with ':' (got: {message[:10]}...)"
            )

        # Match the pattern
        match = cls.CDCL_RESPONSE_PATTERN.match(message)
        if not match:
            raise CDCLParseError(
                f"Response message doesn't match CDCL format: {message}"
            )

        # Extract components
        serial_number, cmd_id, prefix, params_str = match.groups()

        # Parse parameters (whitespace-separated)
        params = params_str.split() if params_str else []

        return CDCLResponse(
            serial_number=serial_number,
            command_id=cmd_id,
            prefix=prefix,
            params=params,
            raw_message=message
        )

    @classmethod
    def parse_command(cls, command: str) -> CDCLCommand:
        """
        Parse a CDCL protocol command

        Args:
            command: Raw CDCL command string

        Returns:
            CDCLCommand object with parsed fields

        Raises:
            CDCLParseError: If command doesn't match CDCL format"""
        if not command:
            raise CDCLParseError("Empty message")

        # Strip whitespace
        message = command.strip()

        match = cls.CDCL_COMMAND_PATTERN.match(command)
        if not match:
            raise CDCLParseError(
                f"Command doesn't match CDCL format: {command}"
            )

        command_prefix, command_id, params_str = match.groups()

        # Parse parameters (whitespace-separated)
        params = params_str.split() if params_str else []

        return CDCLCommand(
            command=command_prefix,
            command_id=command_id,
            params=params,
            raw_message=message
        )

    @classmethod
    def is_cdcl_response(cls, message: str) -> bool:
        """
        Check if a message looks like a CDCL message without full parsing.

        Args:
            message: String to check

        Returns:
            True if message starts with colon and looks like CDCL format
        """
        if not message or not isinstance(message, str):
            return False
        return message.strip().startswith(':')

    @classmethod
    def parse_param_as_int(cls, param: str) -> int:
        """Parse a parameter as integer with error handling"""
        try:
            return int(param)
        except (ValueError, TypeError):
            raise CDCLParseError(f"Cannot parse '{param}' as integer")

    @classmethod
    def parse_param_as_float(cls, param: str) -> float:
        """Parse a parameter as float with error handling"""
        try:
            return float(param)
        except (ValueError, TypeError):
            raise CDCLParseError(f"Cannot parse '{param}' as float")


class PlateReaderModeCalculator:
    """Utility class for manipulating plate reader mode bits"""

    @staticmethod
    def set_led_mode(base_mode: int, led_index: int) -> int:
        """
        Set the LED mode in a mode value.

        Args:
            base_mode: The base mode value to modify
            led_index: The LED index to set (0-7)

        Returns:
            Modified mode value with LED index set
        """
        # Un-set the LED mode field (bits 25-30, 6 bits total)
        base_mode &= ~(0b111111 << 25)

        # Add the new LED index value
        led_val_to_add = (2 ** 25) * led_index
        return base_mode + led_val_to_add

    @staticmethod
    def get_led_index(mode: int) -> int:
        """
        Extract the LED index from a mode value.

        Args:
            mode: Mode value to dissect

        Returns:
            LED index (0-7)
        """
        led_mode = (mode >> 25) - (mode >> 28)
        return led_mode



"""
Experiment results collection and export for plate reader data
"""
from .cdcl import CDCLParser, PlateReaderModeCalculator
from .cdcl.codes import GenericResponsePrefixes
from .utils import index_to_well
from dataclasses import dataclass, field
from typing import List, Dict, Optional, Any, Callable
from datetime import datetime
import json
import csv
import io
from enum import Enum


class DataFormat(Enum):
    """Export data formats"""
    CSV = "csv"
    JSON = "json"
    DICT = "dict"


@dataclass
class DataPoint:
    """Single measurement data point"""
    timestamp: int  # Unix timestamp
    well: str  # Well identifier (e.g., "A1", "B2")
    wavelength: int  # Wavelength in nm
    value: float  # OD or fluorescence reading
    temperature: Optional[float] = None  # Temperature at time of reading
    metadata: Dict[str, Any] = field(default_factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary"""
        return {
            'timestamp': self.timestamp,
            'datetime': datetime.fromtimestamp(self.timestamp).isoformat(),
            'well': self.well,
            'wavelength': self.wavelength,
            'value': self.value,
            'temperature': self.temperature,
            **self.metadata
        }


class ExperimentResults:
    """
    Container for experiment results with export capabilities.

    Collects data points during an experiment and provides methods
    for exporting to various formats.
    """

    def __init__(
        self,
        experiment_id: str,
        experiment_name: str = "",
        wavelengths: Optional[List[int]] = None,
        plate_layout: Optional[Dict[str, Any]] = None
    ):
        """
        Initialize experiment results container.

        Args:
            experiment_id: Unique experiment identifier
            experiment_name: Human-readable experiment name
            wavelengths: List of wavelengths being measured
            plate_layout: Dictionary describing plate layout (rows, cols, etc.)
        """
        self.experiment_id = experiment_id
        self.experiment_name = experiment_name
        self.wavelengths = wavelengths or []
        self.plate_layout = plate_layout or {}

        # Store data points (individual well OD reading)
        self._data_points: List[DataPoint] = []

        # Metadata
        self.start_time: Optional[float] = None
        self.end_time: Optional[float] = None
        self.metadata: Dict[str, Any] = {}

        # Callbacks for real-time processing
        self._data_callbacks: List[Callable[[DataPoint], None]] = []

    def add_data_point(
        self,
        timestamp: int,
        well: str,
        wavelength: int,
        value: float,
        temperature: Optional[float] = None,
        **metadata
    ) -> None:
        """
        Add a single data point.

        Args:
            timestamp: Unix timestamp of measurement
            well: Well identifier
            wavelength: Wavelength in nm
            value: OD or fluorescence value
            temperature: Optional temperature reading
            **metadata: Additional metadata
        """
        data_point = DataPoint(
            timestamp=timestamp,
            well=well,
            wavelength=wavelength,
            value=value,
            temperature=temperature,
            metadata=metadata
        )

        self._data_points.append(data_point)

        # Update start/end times
        if self.start_time is None or timestamp < self.start_time:
            self.start_time = timestamp
        if self.end_time is None or timestamp > self.end_time:
            self.end_time = timestamp

        # Trigger callbacks
        for callback in self._data_callbacks:
            callback(data_point)

    def add_time_point(
        self,
        timestamp: int,
        measurements: Dict[str, float],
        wavelength: int,
        temperature: Optional[float] = None,
        **metadata
    ) -> None:
        """
        Add all measurements from a single time point.

        Args:
            timestamp: Unix timestamp
            measurements: Dictionary of well -> value
            wavelength: Wavelength in nm
            temperature: Optional temperature
            **metadata: Additional metadata
        """

        # Also add individual data points
        for well, value in measurements.items():
            self.add_data_point(
                timestamp=timestamp,
                well=well,
                wavelength=wavelength,
                value=value,
                temperature=temperature,
                **metadata
            )

    def register_callback(self, callback: Callable[[DataPoint], None]) -> None:
        """
        Register a callback for real-time data processing.

        Args:
            callback: Function that takes a DataPoint
        """
        self._data_callbacks.append(callback)

    def get_data_points(
        self,
        well: Optional[str] = None,
        wavelength: Optional[int] = None
    ) -> List[DataPoint]:
        """
        Get filtered data points.

        Args:
            well: Filter by well (None for all)
            wavelength: Filter by wavelength (None for all)

        Returns:
            List of matching data points
        """
        points = self._data_points

        if well is not None:
            points = [p for p in points if p.well == well]

        if wavelength is not None:
            points = [p for p in points if p.wavelength == wavelength]

        return points

    def get_well_timeseries(
        self,
        well: str,
        wavelength: Optional[int] = None
    ) -> List[tuple[float, float]]:
        """
        Get time series data for a specific well.

        Args:
            well: Well identifier
            wavelength: Optional wavelength filter

        Returns:
            List of (timestamp, value) tuples sorted by time
        """
        points = self.get_data_points(well=well, wavelength=wavelength)
        timeseries = [(p.timestamp, p.value) for p in points]
        timeseries.sort(key=lambda x: x[0])
        return timeseries

    def __sort_wells(self, wells):

        def sort_key(well):
            # Split into row (letters) and column (numbers)
            row = well[0].upper()        # e.g. 'A'
            col = int(well[1:])          # e.g. 1, 2, 12
            return (row, col)

        return sorted(wells, key=sort_key)

    def get_wells(self) -> List[str]:
        """Get list of all wells with data"""
        wells = set(p.well for p in self._data_points)
        return self.__sort_wells(wells)

    def get_wavelengths(self) -> List[int]:
        """Get list of all wavelengths measured"""
        wavelengths = set(p.wavelength for p in self._data_points)
        return sorted(wavelengths)

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert results to dictionary.

        Returns:
            Dictionary containing all experiment data
        """
        return {
            'experiment_id': self.experiment_id,
            'experiment_name': self.experiment_name,
            'start_time': self.start_time,
            'end_time': self.end_time,
            'start_datetime': datetime.fromtimestamp(self.start_time).isoformat() if self.start_time else None,
            'end_datetime': datetime.fromtimestamp(self.end_time).isoformat() if self.end_time else None,
            'duration_seconds': (self.end_time - self.start_time) if self.start_time and self.end_time else None,
            'wavelengths': self.wavelengths,
            'plate_layout': self.plate_layout,
            'metadata': self.metadata,
            'data_points': [p.to_dict() for p in self._data_points],
            'total_measurements': len(self._data_points),
            'wells_measured': self.get_wells(),
        }

    def to_json(self, indent: int = 2) -> str:
        """
        Export results as JSON string.

        Args:
            indent: JSON indentation level

        Returns:
            JSON string
        """
        return json.dumps(self.to_dict(), indent=indent)

    def to_csv(
        self,
        format: str = "long",
        include_metadata: bool = True
    ) -> str:
        """
        Export results as CSV string.

        Args:
            format: "long" for one row per measurement,
                   "wide" for one row per timepoint with well columns
            include_metadata: Include temperature and other metadata

        Returns:
            CSV string
        """
        output = io.StringIO()

        if format == "long":
            # Long format: one row per measurement
            writer = csv.writer(output)

            # Header
            header = ['timestamp', 'datetime', 'well', 'wavelength', 'value']
            if include_metadata:
                header.extend(['temperature'])
            writer.writerow(header)

            # Data rows
            for point in self._data_points:
                row = [
                    point.timestamp,
                    datetime.fromtimestamp(point.timestamp).isoformat(),
                    point.well,
                    point.wavelength,
                    point.value
                ]
                if include_metadata:
                    row.append(point.temperature)
                writer.writerow(row)

        elif format == "wide":
            # Wide format: one row per timepoint, columns for each well
            writer = csv.writer(output)

            # Get all wells and wavelengths
            wells = self.get_wells()
            wavelengths = self.get_wavelengths()

            # Group by wavelength
            for wavelength in wavelengths:
                # Header
                header = ['timestamp', 'datetime']
                if include_metadata:
                    header.append('temperature')
                header.extend(wells)
                writer.writerow(header)

                # Group data by timestamp
                time_data: Dict[float, Dict[str, Any]] = {}
                for point in self._data_points:
                    if point.wavelength == wavelength:
                        if point.timestamp not in time_data:
                            time_data[point.timestamp] = {
                                'temperature': point.temperature,
                                'wells': {}
                            }
                        time_data[point.timestamp]['wells'][point.well] = point.value

                # Write rows
                for timestamp in sorted(time_data.keys()):
                    data = time_data[timestamp]
                    row = [
                        timestamp,
                        datetime.fromtimestamp(timestamp).isoformat()
                    ]
                    if include_metadata:
                        row.append(data['temperature'])

                    # Add well values (empty string if missing)
                    for well in wells:
                        row.append(data['wells'].get(well, ''))

                    writer.writerow(row)

                # Blank line between wavelengths
                if len(wavelengths) > 1:
                    writer.writerow([])

        else:
            raise ValueError(f"Unknown format: {format}. Use 'long' or 'wide'")

        return output.getvalue()

    def save_json(self, filepath: str, indent: int = 2) -> None:
        """
        Save results to JSON file.

        Args:
            filepath: Path to output file
            indent: JSON indentation level
        """
        with open(filepath, 'w') as f:
            f.write(self.to_json(indent=indent))

    def save_csv(
        self,
        filepath: str,
        format: str = "long",
        include_metadata: bool = True
    ) -> None:
        """
        Save results to CSV file.

        Args:
            filepath: Path to output file
            format: "long" or "wide"
            include_metadata: Include temperature and other metadata
        """
        with open(filepath, 'w') as f:
            f.write(self.to_csv(format=format, include_metadata=include_metadata))

    def __len__(self) -> int:
        """Return number of data points"""
        return len(self._data_points)

    def __repr__(self) -> str:
        """String representation"""
        duration = ""
        if self.start_time and self.end_time:
            duration_sec = self.end_time - self.start_time
            duration = f", duration={duration_sec:.0f}s"

        return (f"ExperimentResults(id='{self.experiment_id}', "
                f"name='{self.experiment_name}', "
                f"measurements={len(self._data_points)}, "
                f"wells={len(self.get_wells())}{duration})")


class DataTransmissionParser:
    """
    Parser for CDCL data transmission messages.
    """

    @staticmethod
    def parse_data_message(
        message: str,
        results: ExperimentResults
    ) -> None:
        """
        Parse a data transmission message and add to results.

        Args:
            message: Raw CDCL message (typically starts with D prefix)
            results: ExperimentResults object to populate

        """
        response = CDCLParser.parse_response(message)
        if response.prefix == GenericResponsePrefixes.DATA_TRANSMISSION:
            experiment_id = response.params[0].strip().replace('\"', '')
            print(response.params)
            if experiment_id != results.experiment_id:
                print("Data transmission is for a different experiment")
                return results
            timestamp = int(response.params[1])

            if not isinstance(timestamp, int):
                print('timestamp not an integer')
            temperature = response.params[2]
            # mode is the 4th item
            mode = int(response.params[3])
            wavelength_index = PlateReaderModeCalculator.get_led_index(mode)
            wavelength = results.wavelengths[wavelength_index]
            for i in range(4, len(response.params)):
                well = index_to_well(i-4)
                print(f'well {well} value: {response.params[i]}')
                results.add_data_point(
                    timestamp=timestamp,
                    well=well,
                    value=response.params[i],
                    temperature=temperature,
                    wavelength=wavelength
                )

            return results
        else:
            raise ValueError(
                "Response is not a valid Data Transmission message")



import pytest
from cerillo import StratusReader
from cerillo.cdcl import FullExperimentBuilder, ExperimentType


class TestStratusSimulation:
    """Test suite for Stratus reader in simulation mode"""

    def test_initialization(self):
        """Test that Stratus reader initializes correctly in simulation mode"""
        reader = StratusReader(simulate=True)
        reader.connect()
        assert reader.simulate is True
        assert reader.plate_type == "96-well"
        reader.disconnect()

    def test_deprecated_read_absorbance(self):
        """Test that Stratus reader initializes correctly in simulation mode"""
        reader = StratusReader(simulate=True)
        reader.connect()
        assert reader.simulate is True
        assert reader.plate_type == "96-well"
        res = reader.read_absorbance(450, 5, 10)
        print(res)
        reader.disconnect()

    # def test_real_stratus(self):
    #     """Test a complete experiment setup workflow"""
    #     builder = FullExperimentBuilder(command_id="test")

    #     # Configure plate template
    #     builder.plate().set_96_well_plate()
    #     builder.plate().set_manufacturer("Greiner")
    #     builder.plate().set_model("655101")
    #     builder.plate().set_flat_bottom()
    #     builder.plate().set_tissue_culture_treated()

    #     # Configure experiment
    #     builder.experiment().set_name("Growth Curve")
    #     builder.experiment().set_user("scientist@lab.com")
    #     builder.experiment().set_interval(300)
    #     builder.experiment().set_duration(86400)
    #     builder.experiment().set_type(ExperimentType.KINETIC)

    #     messages = builder.build_all_messages()
    #     reader = StratusReader(port="/dev/ttyUSB0")
    #     reader.connect()
    #     reader.start_experiment(builder)


if __name__ == "__main__":
    # Run tests with pytest
    pytest.main([__file__, "-v"])



# src/opentrons_plate_reader/start_commands.py
"""Helper classes for building PLate Reader start commands"""

from dataclasses import dataclass, field
from typing import Optional, List
from .cdcl import CDCLCommand
from .codes import (
    StartCommandPrefixes,
    PlateTemplateMetadataPrefixes,
    PlateTemplatePrefixes,
    PlateTreatmentType,
    WellBottomType,
    PlateLidType,
    ExperimentType
)


@dataclass
class StartCommand(CDCLCommand):
    """Base start command (A prefix)"""
    command: str = field(default="A", init=False)
    command_id: Optional[str] = None
    params: Optional[List[str]] = None

    def __post_init__(self):
        if self.params is None:
            self.params = []
        super().__post_init__()


@dataclass
class ResetStartCommand(StartCommand):
    """Reset start command to defaults (A 0 or A R)"""

    def __post_init__(self):
        self.params = [StartCommandPrefixes.RESET.value]
        super().__post_init__()


@dataclass
class SetIntervalCommand(StartCommand):
    """Set measurement interval in seconds (A i <interval>)"""
    interval: int = 300  # 5 minutes

    def __post_init__(self):
        self.params = [StartCommandPrefixes.INTERVAL.value, str(self.interval)]
        super().__post_init__()


@dataclass
class SetDurationCommand(StartCommand):
    """Set measurement duration in seconds (A d <duration>)"""
    duration: int = -1  # -1 for infinite

    def __post_init__(self):
        self.params = [StartCommandPrefixes.DURATION.value, str(self.duration)]
        super().__post_init__()


@dataclass
class SetModeCommand(StartCommand):
    """Set measurement mode bitfield (A m <mode>)"""
    mode: int = 0

    def __post_init__(self):
        self.params = [StartCommandPrefixes.MODE.value, str(self.mode)]
        super().__post_init__()


@dataclass
class SetExperimentNameCommand(StartCommand):
    """Set experiment name (A n <name>)"""
    name: str = ""  # Empty to auto-generate

    def __post_init__(self):
        if self.name:
            self.params = [StartCommandPrefixes.NAME.value, self.name]
        else:
            self.params = [StartCommandPrefixes.NAME.value]
        super().__post_init__()


@dataclass
class SetExperimentIDCommand(StartCommand):
    """Set experiment ID (A I <ID>)"""
    experiment_id: str = ""  # Empty to auto-generate

    def __post_init__(self):
        if self.experiment_id:
            self.params = [
                StartCommandPrefixes.EXPERIMENT_ID.value, self.experiment_id]
        else:
            self.params = [StartCommandPrefixes.EXPERIMENT_ID.value]
        super().__post_init__()


@dataclass
class SetUserCommand(StartCommand):
    """Set user who started experiment (A U <user>)"""
    user: str = ""

    def __post_init__(self):
        self.params = [StartCommandPrefixes.USER.value, self.user]
        super().__post_init__()


@dataclass
class SetExperimentTypeCommand(StartCommand):
    """Set experiment type (A t <type>)"""
    exp_type: ExperimentType = ExperimentType.ENDPOINT

    def __post_init__(self):
        self.params = [StartCommandPrefixes.TYPE.value,
                       str(self.exp_type.value)]
        super().__post_init__()


@dataclass
class SetQueueCommand(StartCommand):
    """Set whether to queue experiment (A q <0|1>)"""
    queue: bool = False

    def __post_init__(self):
        self.params = [StartCommandPrefixes.QUEUE.value,
                       "1" if self.queue else "0"]
        super().__post_init__()


@dataclass
class ExecuteStartCommand(StartCommand):
    """Execute the start command (A .)"""

    def __post_init__(self):
        self.params = [StartCommandPrefixes.EXECUTE.value]
        super().__post_init__()


# Plate template commands

@dataclass
class PlateTemplateCommand(CDCLCommand):
    """Base plate template command (A P prefix)"""
    command: str = field(default="A", init=False)
    command_id: Optional[str] = None
    params: Optional[List[str]] = None

    def __post_init__(self):
        if self.params is None:
            self.params = []
        super().__post_init__()


@dataclass
class SetPlateRowsCommand(PlateTemplateCommand):
    """Set number of rows (A P M R <rows>)"""
    rows: int = 8

    def __post_init__(self):
        self.params = [
            StartCommandPrefixes.PLATE.value + PlateTemplatePrefixes.METADATA.value +
            PlateTemplateMetadataPrefixes.ROWS.value,
            str(self.rows)
        ]
        super().__post_init__()


@dataclass
class SetPlateColsCommand(PlateTemplateCommand):
    """Set number of columns (A P M C <cols>)"""
    cols: int = 12

    def __post_init__(self):
        self.params = [
            StartCommandPrefixes.PLATE.value + PlateTemplatePrefixes.METADATA.value +
            PlateTemplateMetadataPrefixes.COLS.value,
            str(self.cols)
        ]
        super().__post_init__()


@dataclass
class SetWellDiameterCommand(PlateTemplateCommand):
    """Set well diameter in microns (A P M D <diameter>)"""
    diameter_um: int = 6400  # Default for 96-well

    def __post_init__(self):
        self.params = [
            StartCommandPrefixes.PLATE.value + PlateTemplatePrefixes.METADATA.value +
            PlateTemplateMetadataPrefixes.DIAMETER.value,
            str(self.diameter_um)
        ]
        super().__post_init__()


@dataclass
class SetWellHeightCommand(PlateTemplateCommand):
    """Set well height in microns (A P M H <height>)"""
    height_um: int = 10900  # Default for 96-well

    def __post_init__(self):
        self.params = [
            StartCommandPrefixes.PLATE.value + PlateTemplatePrefixes.METADATA.value +
            PlateTemplateMetadataPrefixes.HEIGHT.value,
            str(self.height_um)
        ]
        super().__post_init__()


@dataclass
class SetWellVolumeCommand(PlateTemplateCommand):
    """Set well volume in microliters (A P M V <volume>)"""
    volume_ul: int = 360  # Default for 96-well

    def __post_init__(self):
        self.params = [
            StartCommandPrefixes.PLATE.value + PlateTemplatePrefixes.METADATA.value +
            PlateTemplateMetadataPrefixes.VOLUME.value,
            str(self.volume_ul)
        ]
        super().__post_init__()


@dataclass
class SetPlateTreatmentCommand(PlateTemplateCommand):
    """Set plate treatment type (A P M T <type>)"""
    treatment: PlateTreatmentType = PlateTreatmentType.UNKNOWN

    def __post_init__(self):
        self.params = [
            StartCommandPrefixes.PLATE.value + PlateTemplatePrefixes.METADATA.value +
            PlateTemplateMetadataPrefixes.TREATMENT.value,
            self.treatment.value
        ]
        super().__post_init__()


@dataclass
class SetWellBottomCommand(PlateTemplateCommand):
    """Set well bottom type (A P M B <type>)"""
    bottom: WellBottomType = WellBottomType.UNKNOWN

    def __post_init__(self):
        self.params = [
            StartCommandPrefixes.PLATE.value + PlateTemplatePrefixes.METADATA.value +
            PlateTemplateMetadataPrefixes.BOTTOM.value,
            self.bottom.value
        ]
        super().__post_init__()


@dataclass
class SetPlateLidCommand(PlateTemplateCommand):
    """Set plate lid type (A P M L <type>)"""
    lid: PlateLidType = PlateLidType.UNKNOWN

    def __post_init__(self):
        self.params = [
            StartCommandPrefixes.PLATE.value + PlateTemplatePrefixes.METADATA.value +
            PlateTemplateMetadataPrefixes.LID.value,
            self.lid.value
        ]
        super().__post_init__()


@dataclass
class AssignDetectorToWellCommand(PlateTemplateCommand):
    """Assign detector to well (A P M W <detector> <well>)"""
    detector_num: int = 0
    well_num: int = 0

    def __post_init__(self):
        self.params = [
            StartCommandPrefixes.PLATE.value + PlateTemplatePrefixes.METADATA.value +
            PlateTemplateMetadataPrefixes.WELL_DETECTOR.value,
            str(self.detector_num),
            str(self.well_num)
        ]
        super().__post_init__()


@dataclass
class SetPlateManufacturerCommand(PlateTemplateCommand):
    """Set plate manufacturer (A P M M <manufacturer>)"""
    manufacturer: str = ""

    def __post_init__(self):
        self.params = [
            StartCommandPrefixes.PLATE.value + PlateTemplatePrefixes.METADATA.value +
            PlateTemplateMetadataPrefixes.MANUFACTURER.value,
            self.manufacturer
        ]
        super().__post_init__()


@dataclass
class SetPlateModelCommand(PlateTemplateCommand):
    """Set plate model number (A P M m <model>)"""
    model: str = ""

    def __post_init__(self):
        self.params = [
            StartCommandPrefixes.PLATE.value + PlateTemplatePrefixes.METADATA.value +
            PlateTemplateMetadataPrefixes.MODEL.value,
            self.model
        ]
        super().__post_init__()


@dataclass
class SetPlateTemplateNameCommand(PlateTemplateCommand):
    """Set plate template name (A P M n <name>)"""
    name: str = ""

    def __post_init__(self):
        self.params = [
            StartCommandPrefixes.PLATE.value + PlateTemplatePrefixes.METADATA.value +
            PlateTemplateMetadataPrefixes.TEMPLATE_NAME.value,
            self.name
        ]
        super().__post_init__()


@dataclass
class AddPlateLayoutLineCommand(PlateTemplateCommand):
    """Add to plate layout file (A P G <line>)"""
    line: str = ""

    def __post_init__(self):
        self.params = [
            PlateTemplateMetadataPrefixes.LAYOUT.value,
            self.line
        ]
        super().__post_init__()



from abc import ABC, abstractmethod
from ast import Lambda
from deprecated import deprecated
from typing import Callable, Optional, Dict, Any, List
import serial
import string
import random
import time
from .cdcl.cdcl import CDCLCommand, CDCLParser, CDCLResponse, PlateReaderModeCalculator
from .experiment_results import DataTransmissionParser, ExperimentResults
from .cdcl.experiment_builder import FullExperimentBuilder
from .cdcl.codes import ExperimentMetadataCodes, PlateReaderSpecificInstructionPrefixes, GenericInstructionPrefixes, StartCommandPrefixes, GenericResponsePrefixes, ExperimentType, PlateReaderSpecificResponsePrefixes
from operator import attrgetter


class CerilloBasePlateReader(ABC):
    """Base class for all Cerillo plate readers"""

    def __init__(self, port: str = '/dev/ttyUSB0', baudrate: int = 250000,
                 timeout: float = 120.0, simulate: bool = False):
        self.port = port
        self.baudrate = baudrate
        self.timeout = timeout
        self.simulate = simulate
        self._connection: Optional[serial.Serial] = None
        self.serial_number = None
        self.model_number = None
        self.name = None
        self.status = None
        self.firmware_version = None
        self.wavelengths: List[int] = []
        if simulate:
            self.wavelengths = [450]

    def __repr__(self):
        """Developer-friendly representation"""
        conn_status = "connected" if self._connection and self._connection.is_open else "disconnected"
        sim_str = " [SIMULATION]" if self.simulate else ""

        # Basic info always shown
        parts = [f"{self.__class__.__name__}("]

        # Device info if available
        if self.name:
            parts.append(f"name='{self.name}'")
        if self.serial_number:
            parts.append(f"serial={self.serial_number}")
        if self.model_number:
            parts.append(f"model='{self.model_number}'")
        parts.append(f"wavelengths={self.wavelengths}")

        # Connection info
        parts.append(f"port='{self.port}'")
        parts.append(f"status={conn_status}{sim_str}")
        return ", ".join(parts) + ")"

    def connect(self) -> None:
        """Establish connection to the plate reader"""
        if self.simulate:
            print(f"[SIMULATION] Connected to {self.__class__.__name__}")
            self.serial_number = "12345"
            return

        try:
            self._connection = serial.Serial(
                self.port,
                self.baudrate,
                timeout=self.timeout
            )

            start_time = time.time()
            while (self.serial_number == None):
                # Check if timeout exceeded
                elapsed = time.time() - start_time
                if elapsed > self.timeout:
                    raise TimeoutError(
                        f"Device startup timed out after {self.timeout} seconds waiting for serial number line")

                line = self._connection.readline().decode().strip()
                print(line)
                if (line.startswith(":")):
                    self.serial_number = line[1:line.find(" ")]
                    print(
                        f"Connected to device with serial number {self.serial_number}")

            self.initialize()
        except serial.SerialException as e:
            raise ConnectionError(f"Failed to connect to plate reader: {e}")

    def disconnect(self) -> None:
        """Close connection to the plate reader"""
        if self.simulate:
            print(f"[SIMULATION] Disconnected from {self.__class__.__name__}")
            return

        if self._connection and self._connection.is_open:
            self._connection.close()

    def send_command(self, command: CDCLCommand, timeout: int = None, onNewLineCallback: Callable[[str], None] = None) -> CDCLResponse:
        """
        Send a CDCL command to the connected device and wait for a response.

        Description:
            This method will ensure a connection is present (calling `connect()` if
            necessary), write a properly formatted command, then block until a
            response matching the command id is received. When in simulation mode
            the function returns a synthetic `CDCLResponse`.

        Usage Note:
            Prefer to use higher-level plate reader control commands like `start_endpoint_experiment()`
            instead of this function directly

        Args:
            command: CDCLCommand object to send
            timeout: Per-command timeout in seconds. If None, the instance's
                `timeout` property is used.
            onNewLineCallback: Optional callback function invoked for every
                newline received from the device. The callback receives the
                decoded line as a single string parameter.

        Returns:
            CDCLResponse: Parsed CDCL response matching the sent command id.

        Raises:
            TimeoutError: If the response is not received within `timeout`.
            ConnectionError: When there is a failure opening the serial port.
            TypeError: If the received message isn't a valid CDCL response.
        """
        # set timeout to classes if not passed
        if timeout == None:
            timeout = self.timeout
        if self.simulate:
            return CDCLResponse(
                serial_number=self.serial_number,
                command_id=command.command_id,
                prefix=command.command,
                params=command.params,
                raw_message=f':{self.serial_number} {command.command}*{command.command_id} {command.params}'
            )
        try:

            # Add command ID if needed
            command_id = command.command_id
            if (not command.command_id or command.command_id == ""):
                command_id = ''.join(random.choice(
                    string.ascii_uppercase + string.digits) for _ in range(4))
                command.command_id = command_id
            # Send command (adjust formatting for your device)
            device_command = f"{command.to_message()}\n".encode()
            self.log_serial(direction="send", message=device_command)

            # return early if simulating
            if self.simulate:
                return

            if not self._connection or not self._connection.is_open:
                self.connect()

            # Clear any buffered input before sending command
            self._connection.reset_input_buffer()

            self._connection.write(device_command)
            self._connection.flush()
            # Read response (adjust based on your protocol)
            # clear echo
            response = ''
            echo = 0
            start_time = time.time()
            while (response == ''):
                # Check if timeout exceeded
                elapsed = time.time() - start_time
                if elapsed > timeout:
                    raise TimeoutError(
                        f"Command {command} timed out after {timeout} seconds waiting for response with ID {command_id}")

                line = self._connection.readline().decode().strip()
                self.log_serial(direction="receive", message=line)
                # handle callback
                if onNewLineCallback:
                    onNewLineCallback(line)
                # print(line)
                if (command_id in line):
                    # clear the echo
                    if (not echo):
                        echo = 1
                    else:
                        response = line
            print(f"$Command: {command} | $Response: {response}")

            if not CDCLParser.is_cdcl_response(response):
                raise TypeError(
                    "response is not a valid CDCL response", response)
            response = CDCLParser.parse_response(response)

            return response
        except Exception as e:
            print(f"Command failed: {e}")
            raise

    def log_serial(self, direction: str, message: str, max_width: int = 80):
        """
        Log a serial communication event to stdout.

        The helper prints a single line prefixed with 'TX' or 'RX' and an arrow
        to show direction. The `simulate` flag is included in the printed
        information when set. This is intended for developer debugging, not
        structured logging.

        Args:
            direction: "send" for TX entries or any other string for RX.
            message: Message that was transmitted or received.
            max_width: Optional width to enforce truncation for readability.
        """
        if direction == "send":
            arrow = "→"
            print(f'{"[SIMULATION] " if self.simulate else ""}TX {arrow} {message}')
        else:  # receive
            arrow = "←"
            print(f'{"[SIMULATION] " if self.simulate else ""}RX {arrow} {message}')

    def initialize(self) -> None:
        """
        Initialize metadata on the connected plate reader.

        This method queries the device for info (`GenericInstructionPrefixes.GET_INFO`),
        status, and the available wavelengths, then fills in attributes like
        `name`, `model_number`, `firmware_version`, `status`, and
        `wavelengths` on the instance. In simulation mode the method is a no-op
        and simply prints a message.
        """
        if self.simulate:
            print("[SIMULATION] Plate Reader initialized")
            return
        # TODO get serial number
        info_command = CDCLCommand(
            command=GenericInstructionPrefixes.GET_INFO)
        info_response = self.send_command(info_command, timeout=60)
        print(info_response)
        # find the param that starts with "n" and slice the prefix
        self.name = next(
            x for x in info_response.params if x.startswith("n"))[1:]
        # find the param that starts with "n" and slice the prefix
        self.model_number = next(
            x for x in info_response.params if x.startswith("m"))[1:]
        self.firmware_version = next(
            x for x in info_response.params if x.startswith("f"))[1:]

        status_command = CDCLCommand(
            command=GenericInstructionPrefixes.GET_STATUS)
        status_response = self.send_command(status_command)
        self.status = status_response.params[0]

        wavelength_command = CDCLCommand(
            command=PlateReaderSpecificInstructionPrefixes.GET_WAVELENGTHS)
        wavelength_response = self.send_command(wavelength_command)
        if wavelength_response.params:
            self.wavelengths = list(map(int, wavelength_response.params))

        print(self)

    def calibrate(self, wavelength: int) -> CDCLResponse:
        """
        Calibrate the plate reader for a specific wavelength.

        Args:
            wavelength: The wavelength in nm to calibrate (must be present in
                `self.wavelengths`).

        Returns:
            CDCLResponse object for the calibrate command response.

        Raises:
            ValueError: If the provided wavelength is not supported by device.
        """
        wavelength_index = self.wavelengths.index(wavelength)
        if wavelength_index == -1:
            raise ValueError(f'wavelength {wavelength} not found on device')
        mode = PlateReaderModeCalculator.set_led_mode(0, wavelength_index)
        calibrate_command = CDCLCommand(
            command=PlateReaderSpecificInstructionPrefixes.CALIBRATE, params=[str(mode)])
        return self.send_command(calibrate_command)

    @deprecated(version="0.0.6", reason="Use start_kinetic_experiment() or start_endpoint_experiment() instead")
    def read_absorbance(self, wavelength: int, interval: int, duration: int,  **kwargs) -> Dict[str, Any]:
        """
          @Deprecated as of v0.0.6 - Use start_kinetic_experiment() or start_endpoint_experiment() instead
          Start an endpoint experiment for a given wavelength, reading every interval for a set duration

          Args:
              wavelength: (nm) must be one of the available wavelengths on device, otherwise a ValueError will be thrown
              interval: (seconds) Time in between start of each reading
              duration: (hours) How long the experiment will continue for

          """
        if self.simulate:
            print("[SIMULATION] Read absorbance")

        return self.start_kinetic_experiment(wavelength, interval, duration)

    def start_experiment(self, builder: FullExperimentBuilder) -> Dict[str, any]:
        """
        Send a complete experiment configuration to the device and gather results.

        The function constructs CDCL commands via the provided
        `FullExperimentBuilder`, sends them to the device and blocks until at
        least one data point is received. It returns the constructed
        `ExperimentResults` object.

        Args:
            builder: A `FullExperimentBuilder` instance containing both an
                experiment configuration and plate template.

        Returns:
            ExperimentResults: A populated experiment results container.
        """
        commands = builder.build_all()
        if self.simulate:
            return {}
        # container to store results
        results = None
        for command in commands:
            def onNewDataCallback(line):
                nonlocal results   # <--- allow assignment to outer 'results'
                if CDCLParser.is_cdcl_response(line):
                    response = CDCLParser.parse_response(line)

                    if (response.prefix == GenericResponsePrefixes.GET_EXPERIMENT and response.params[0].startswith('m')):
                        mode = None
                        # loop through metadata (excluding first one which is experiment id)
                        for i in range(1, len(response.params)):
                            if response.params[i].startswith(ExperimentMetadataCodes.MODE.value):
                                mode = response.params[i][1:]
                        wavelength = None
                        if mode:
                            led_index = PlateReaderModeCalculator.get_led_index(
                                int(mode))
                            wavelength = self.wavelengths[led_index]
                        results = ExperimentResults(
                            # strip off the "m"
                            experiment_id=response.params[0][1:],
                            experiment_name=response.params[1],
                            wavelengths=[wavelength],
                        )
                        print("results is defined")

            self.send_command(
                command=command, timeout=None, onNewLineCallback=onNewDataCallback)

        # TODO wait for end of experiment instead of just one data point
        while (len(results.get_data_points()) == 0):
            line = self._connection.readline().decode().strip()
            self.log_serial(direction="receive", message=line)
            if CDCLParser.is_cdcl_response(line):
                response = CDCLParser.parse_response(line)
                # look for the metadata response message
                print(response)
                print(results)
                if (response.prefix == GenericResponsePrefixes.DATA_TRANSMISSION.value):
                    DataTransmissionParser.parse_data_message(line, results)

        # Export examples
        print("\n=== JSON Export ===")
        print(results.to_json())
        return results

    def start_endpoint_experiment(self, wavelength):
        """
        Start an endpoint experiment for a given wavelength

        Args:
            wavelength: (nm) must be one of the available wavelengths on device, otherwise a ValueError will be thrown
        """
        # get wavelength
        mode = self.__convert_wavelength_to_mode(wavelength)
        builder = FullExperimentBuilder()
        builder.experiment().set_type(ExperimentType.ENDPOINT).set_mode(mode)
        return self.start_experiment(builder)

    def start_kinetic_experiment(self, wavelength, interval=300, duration=None):
        """
        Start a kinetic experiment for a given wavelength

        Args:
            wavelength: (nm) must be one of the available wavelengths on device, otherwise a ValueError will be thrown
            interval (seconds): Defaults to 300 seconds (5 minutes)
        """
        mode = self.__convert_wavelength_to_mode(wavelength)
        builder = FullExperimentBuilder()
        builder.experiment().set_type(ExperimentType.KINETIC).set_mode(
            mode).set_interval(interval)
        if duration:
            builder.experiment().set_duration(duration)
        return self.start_experiment(builder)

    def __convert_wavelength_to_mode(self, wavelength) -> int:
        """
        Takes in a wavelength (in nm) and spits out the mode associated with that wavelength
        If the wavelength is not present on the device a ValueError will be thrown
        """
        wavelength_index = self.wavelengths.index(wavelength)
        if wavelength_index == -1:
            raise ValueError(f'wavelength {wavelength} not found on device')
        # TODO set mode to proper for endpoint or kinetic?
        mode = PlateReaderModeCalculator.set_led_mode(983040, wavelength_index)
        return mode



"""
CDCL Protocol Command and Response Prefixes

These enums serve as the source-of-truth for what prefixes to use for specific
commands and what prefix to expect in response.
"""

from typing import Union
from enum import Enum

# ----------------- Instructions ---------------------


class GenericInstructionPrefixes(str, Enum):
    """Generic instruction prefixes for all devices"""
    GET_CHILDREN = 'h'
    GET_FILE = 'F'
    GET_INFO = 'i'
    GET_QC_MODE_KEY = 'k'
    GET_STATUS = 'x'
    GET_TIMESTAMP = 'Z'
    LIST_FILES = 'f'
    READ_XBEE_BUFFER = 'w'
    REBOOT = '0'
    SET_DEMO_MODE = 'd'
    SET_MODEL_NUMBER = 'N'
    SET_SERIAL_NUMBER = 'n'
    SET_TIMESTAMP = 'z'
    TURN_QC_MODE_OFF = 'q'
    TURN_QC_MODE_ON = 'Q'
    UNLOCK_DEMO = 'u'
    GET_SETTINGS = 'G'
    SET_SETTINGS = 'g'
    LIST_EXPERIMENTS = 'e'
    GET_EXPERIMENT = 'E'
    GET_STORAGE_MODIFIED_TIMESTAMP = 'm'


class PlateReaderSpecificInstructionPrefixes(str, Enum):
    """Plate reader specific instruction prefixes"""
    CALIBRATE = 'c'
    BLINK = 'b'
    READ_PLATE = 'R'
    READ_WELL = 'r'
    SET_STATUS_LED = 'L'
    SET_LED = 'l'
    START = 'S'
    STOP = 's'
    GET_TEMPERATURE = 't'
    EXPERIMENT_SETUP = 'A'
    SET_WAVELENGTHS = 'v'
    GET_WAVELENGTHS = 'V'
    GET_IREF = 'J'
    SET_IREF = 'j'


class CanopySpecificInstructionPrefixes(str, Enum):
    """Canopy specific instruction prefixes"""
    STRESS_MODE_ON = 'M'
    STRESS_MODE_OFF = 'm'
    UPDATE_AUTH = 'U'


# Type aliases for instruction prefix unions
PlateReaderInstructionPrefixes = Union[
    GenericInstructionPrefixes,
    PlateReaderSpecificInstructionPrefixes
]


# ---------------- Responses -----------------------

class GenericResponsePrefixes(str, Enum):
    """Generic response prefixes for all devices"""
    CHILDREN = 'h'
    DEMO_REQ = 'u'
    FILE_LIST = 'f'
    GET_QC_MODE_KEY = 'K'
    INFO = 'i'
    SETTINGS = 'gP'
    STATUS = 'x'
    TEMPERATURE = 't'
    TIMESTAMP = 'Z'
    TURN_QC_MODE_OFF = 'xR'
    TURN_QC_MODE_ON = 'xQ'
    GET_FILE = 'F'
    LIST_EXPERIMENTS = 'e'
    GET_EXPERIMENT = 'E'
    DATA_TRANSMISSION = 'D'
    CMD_ERROR = '!'
    GET_STORAGE_MODIFIED_TIMESTAMP = 'm'
    FILE_STORAGE_OPERATION = 'f'
    BUTTON_CLICK = 'b'
    STOP = 's'


class PlateReaderSpecificResponsePrefixes(str, Enum):
    """Plate reader specific response prefixes"""
    READ_WELL = 'd'
    LEGACY_START = 'S'
    EXPERIMENT_SETUP = 'A'
    SET_WAVELENGTHS = 'v'
    CALIBRATE = 'c'
    IREF = 'j'


# Type aliases for response prefix unions
StratusResponsePrefixes = Union[
    GenericResponsePrefixes,
    PlateReaderSpecificResponsePrefixes
]


class StartCommandPrefixes(str, Enum):
    """Start command (A) sub-prefixes"""
    RESET = '0'  # Reset start command to defaults
    INTERVAL = 'i'  # Set measurement interval (seconds)
    DURATION = 'd'  # Set measurement duration (seconds)
    MODE = 'm'  # Set measurement mode (bitfield)
    NAME = 'n'  # Set experiment name
    EXPERIMENT_ID = 'I'  # Set experiment ID
    USER = 'U'  # Set user who started experiment
    TYPE = 't'  # Set experiment type
    QUEUE = 'q'  # Queue experiment instead of starting
    PLATE = 'P'
    EXECUTE = '.'  # Execute the start command


class PlateTemplateMetadataPrefixes(str, Enum):
    """Plate template metadata (A P M) sub-prefixes"""
    ROWS = 'R'  # Number of rows
    COLS = 'C'  # Number of columns
    DIAMETER = 'D'  # Well diameter (μm)
    HEIGHT = 'H'  # Well height (μm)
    VOLUME = 'V'  # Well volume (μL)
    TREATMENT = 'T'  # Plate treatment type
    BOTTOM = 'B'  # Well-bottom type
    LID = 'L'  # Plate lid type
    WELL_DETECTOR = 'W'  # Assign detector to well
    MANUFACTURER = 'M'  # Plate manufacturer
    MODEL = 'm'  # Plate model number
    TEMPLATE_NAME = 'n'  # Template name
    LAYOUT = 'G'  # plate layout file (the .plate file)


class PlateTemplatePrefixes(str, Enum):
    """Plate template (A P) sub-prefixes"""
    METADATA = 'M'  # Plate metadata commands
    LAYOUT = 'G'  # Add line to layout file
    WELLS = 'W'  # Add line to wells file (deprecated as of v6.6.0)


class PlateTreatmentType(str, Enum):
    """Plate treatment types"""
    UNKNOWN = ''  # Unknown/unspecified
    UNTREATED = 'u'  # Untreated
    TISSUE_CULTURE = 't'  # Tissue-culture treated


class WellBottomType(str, Enum):
    """Well bottom types"""
    UNKNOWN = ''  # Unknown/unspecified
    ROUND = 'r'  # Round bottom
    FLAT = 'f'  # Flat bottom


class PlateLidType(str, Enum):
    """Plate lid types"""
    UNKNOWN = ''  # Unknown/unspecified
    NONE = 'n'  # No lid
    LID = 'l'  # Standard lid
    PERMEABLE_MEMBRANE = 'm'  # Permeable membrane (e.g., Breathe-Easy)
    IMPERMEABLE_MEMBRANE = 'M'  # Impermeable membrane (e.g., PCR plastic)


class ExperimentType(Enum):
    """Experiment types"""
    DEFER_TO_DURATION = -1  # Let duration determine type
    ENDPOINT = 0  # Endpoint measurement
    KINETIC = 1  # Kinetic measurement
    SMART_ENDPOINT = 2  # Smart endpoint (v6.11.0+)
    IREF_SWEEP = 3  # IREF sweep 0-255 by 5 (v6.11.0+)


class ExperimentMetadataCodes(Enum):
    """Experiment Metadata Codes"""
    SERIAL_NUMBER = 'N'     # N-Serial number of the device that ran the experiment
    EXPERIMENT_ID = 'I'     # I	Experiment ID as recorded in the file
    MODE = 'm'              # m	Experiment mode
    MODE_STR = 'M'          # M	Experiment mode descriptive string
    DATA_SCHEMA = 'd'       # d	Data schema
    NAME = 'n'              # n	Experiment name
    INTERVAL = 'i'          # i	Measurement interval
    START_TIMESTAMP = 'S'   # S	Experiment start UNIX timestamp
    SCHEDULED_END_TIMESTAMP = 's'   # s	Experiment scheduled end UNIX timestamp
    USER = 'U'              # U	User who started the experiment
    LENGTH = 'l'            # l	Experiment length; number of data points
    FILE_SIZE = 'f'         # f	Experiment OD file size; bytes
    TYPE = 't'              # t	Experiment type
    KEY_MEASUREMENT = 'k'     # k	Experiment key measurement  #



from typing import Dict, Any, List, Optional
from .base_plate_reader import CerilloBasePlateReader


class StratusReader(CerilloBasePlateReader):
    """Stratus plate reader implementation"""

    def __init__(self, port: str = '/dev/ttyUSB0', **kwargs):
        """
        Initialize a StratusReader instance.

        Args:
            port: Serial port path for the device.
            **kwargs: Additional kwargs forwarded to `CerilloBasePlateReader`.
        """
        super().__init__(port, baudrate=250000, **kwargs)
        self.plate_type = "96-well"
